﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Analysis;
using HeuristicLab.Analysis.QualityAnalysis;
using HeuristicLab.Clients.OKB.RunCreation;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;
using HeuristicLab.Core;
using HeuristicLab.Core.Views;
using HeuristicLab.Data;
using HeuristicLab.Data.Views;
using HeuristicLab.MainForm;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Views;
using HeuristicLab.OptimizationExpertSystem.Common;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.Windows.Forms.DataVisualization.Charting;
using DoubleValue = HeuristicLab.Data.DoubleValue;

namespace HeuristicLab.OptimizationExpertSystem {
  [View("Expert-system Optimizer View")]
  [Content(typeof(ExpertSystem), IsDefaultView = true)]
  public partial class ExpertSystemView : NamedItemView {
    private EnumValueView<SeedingStrategyTypes> seedingStrategyView;
    private CheckedItemListView<IScope> seedingSolutionsView;
    protected virtual bool SuppressEvents { get; set; }
    private bool okbDownloadInProgress;

    public new ExpertSystem Content {
      get { return (ExpertSystem)base.Content; }
      set { base.Content = value; }
    }

    public ExpertSystemView() {
      InitializeComponent();
      // brings progress panel to front (it is not visible by default, but obstructs other elements in designer)
      this.Controls.SetChildIndex(this.progressPanel, 0);
      algorithmStartButton.Text = string.Empty;
      algorithmStartButton.Image = VSImageLibrary.Play;
      algorithmCloneButton.Text = string.Empty;
      algorithmCloneButton.Image = VSImageLibrary.Clone;
      refreshMapButton.Text = string.Empty;
      refreshMapButton.Image = VSImageLibrary.Refresh;
      seedingStrategyView = new EnumValueView<SeedingStrategyTypes>() {
        Dock = DockStyle.Fill
      };
      seedingSolutionsView = new CheckedItemListView<IScope>() {
        Dock = DockStyle.Fill
      };
      solutionsSeedingTableLayoutPanel.SuspendLayout();
      solutionsSeedingTableLayoutPanel.Controls.Add(seedingStrategyView, 0, 0);
      solutionsSeedingTableLayoutPanel.Controls.Add(seedingSolutionsView, 0, 1);
      solutionsSeedingTableLayoutPanel.ResumeLayout(true);
    }

    #region Event Registration
    protected override void DeregisterContentEvents() {
      Content.PropertyChanged -= ContentOnPropertyChanged;
      Content.SuggestedInstances.CollectionReset -= SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsAdded -= SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsMoved -= SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsRemoved -= SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsReplaced -= SuggestedInstancesOnChanged;
      DeregisterProblemEvents(Content.Problem);
      base.DeregisterContentEvents();
    }

    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.PropertyChanged += ContentOnPropertyChanged;
      Content.SuggestedInstances.CollectionReset += SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsAdded += SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsMoved += SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsRemoved += SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsReplaced += SuggestedInstancesOnChanged;
      RegisterProblemEvents(Content.Problem);
    }

    private void DeregisterProblemEvents(OKBProblem problem) {
      if (problem == null) return;
      problem.Solutions.ItemsAdded -= SolutionsOnChanged;
      problem.Solutions.ItemsRemoved -= SolutionsOnChanged;
      problem.Solutions.ItemsReplaced -= SolutionsOnChanged;
      problem.Solutions.CollectionReset -= SolutionsOnChanged;
      problem.ProblemChanged -= ContentOnProblemChanged;
    }

    private void RegisterProblemEvents(OKBProblem problem) {
      if (problem == null) return;
      problem.Solutions.ItemsAdded += SolutionsOnChanged;
      problem.Solutions.ItemsRemoved += SolutionsOnChanged;
      problem.Solutions.ItemsReplaced += SolutionsOnChanged;
      problem.Solutions.CollectionReset += SolutionsOnChanged;
      problem.ProblemChanged += ContentOnProblemChanged;
    }
    #endregion

    protected override void OnContentChanged() {
      base.OnContentChanged();
      SuppressEvents = true;
      okbDownloadInProgress = false;
      try {
        if (Content == null) {
          maxEvaluationsTextBox.Text = String.Empty;
          problemViewHost.Content = null;
          solverParametersView.Content = null;
          solverResultsView.Content = null;
          runsView.Content = null;
          kbViewHost.Content = null;
          problemInstancesView.Content = null;
          seedingStrategyView.Content = null;
          seedingSolutionsView.Content = null;
        } else {
          maxEvaluationsTextBox.Text = Content.MaximumEvaluations.ToString();
          problemViewHost.Content = Content.Problem;
          runsView.Content = Content.Runs;
          kbViewHost.ViewType = typeof(RunCollectionRLDView);
          kbViewHost.Content = Content.KnowledgeBase;
          problemInstancesView.Content = Content.ProblemInstances;
          solverResultsView.Content = Content.CurrentResult;
          seedingStrategyView.Content = Content.SeedingStrategy;
          seedingSolutionsView.Content = Content.SolutionSeedingPool;
        }
      } finally { SuppressEvents = false; }
      UpdateSuggestedInstancesCombobox();
      UpdateSimilarityCalculators();
      UpdateNamesComboboxes();
    }

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      maxEvaluationsTextBox.Enabled = Content != null && !ReadOnly && !Locked;
      problemViewHost.Enabled = Content != null && !ReadOnly && !Locked && !okbDownloadInProgress;
      suggestedInstancesComboBox.Enabled = Content != null && !ReadOnly && !Locked && !okbDownloadInProgress;
      algorithmStartButton.Enabled = Content != null && !ReadOnly && !Locked && suggestedInstancesComboBox.SelectedIndex >= 0;
      algorithmCloneButton.Enabled = Content != null && !ReadOnly && !Locked && suggestedInstancesComboBox.SelectedIndex >= 0;
      runsView.Enabled = Content != null;
      kbViewHost.Enabled = Content != null && !okbDownloadInProgress;
      problemInstancesView.Enabled = Content != null && !okbDownloadInProgress;
      refreshMapButton.Enabled = Content != null;
      okbDownloadButton.Enabled = Content != null && Content.Problem != null && Content.Problem.ProblemId >= 0 && !ReadOnly && !Locked && !okbDownloadInProgress;
    }

    #region Event Handlers
    #region Content events
    private void ContentOnProblemChanged(object sender, EventArgs eventArgs) {
      UpdateSuggestedInstancesCombobox();
      UpdateSimilarityCalculators();
      SetEnabledStateOfControls();
    }

    private void ContentOnPropertyChanged(object sender, PropertyChangedEventArgs e) {
      if (InvokeRequired) {
        Invoke((Action<object, PropertyChangedEventArgs>)ContentOnPropertyChanged, sender, e);
        return;
      }
      SuppressEvents = true;
      try {
        switch (e.PropertyName) {
          case "KnowledgeBase": kbViewHost.Content = Content.KnowledgeBase; break;
          case "MaximumEvaluations": maxEvaluationsTextBox.Text = Content.MaximumEvaluations.ToString(); break;
          case "Problem":
            DeregisterProblemEvents(problemViewHost.Content as OKBProblem);
            problemViewHost.Content = Content.Problem;
            RegisterProblemEvents(Content.Problem);
            UpdateNamesComboboxes();
            break;
          case "ProblemInstances": problemInstancesView.Content = Content.ProblemInstances; break;
          case "CurrentResult": solverResultsView.Content = Content.CurrentResult; break;
        }
      } finally { SuppressEvents = false; }
      SetEnabledStateOfControls();
    }

    private void SuggestedInstancesOnChanged(object sender, EventArgs e) {
      UpdateSuggestedInstancesCombobox();
    }

    private void SolutionsOnChanged(object sender, EventArgs e) {
      UpdateNamesComboboxes();
      UpdateSolutionVisualization();
    }
    #endregion

    #region Control events
    private void MaxEvaluationsTextBoxOnValidating(object sender, CancelEventArgs e) {
      if (SuppressEvents) return;
      if (InvokeRequired) {
        Invoke((Action<object, CancelEventArgs>)MaxEvaluationsTextBoxOnValidating, sender, e);
        return;
      }
      int value;
      if (!int.TryParse(maxEvaluationsTextBox.Text, out value)) {
        e.Cancel = !maxEvaluationsTextBox.ReadOnly && maxEvaluationsTextBox.Enabled;
        errorProvider.SetError(maxEvaluationsTextBox, "Please enter a valid integer number.");
      } else {
        Content.MaximumEvaluations = value;
        e.Cancel = false;
        errorProvider.SetError(maxEvaluationsTextBox, null);
      }
    }

    private void RefreshMapButtonOnClick(object sender, EventArgs e) {
      Content.UpdateInstanceProjection();
      UpdateProjectionComboBox();
    }

    private void OkbDownloadButtonOnClick(object sender, EventArgs e) {
      if (Content.Problem.ProblemId == -1) {
        MessageBox.Show("Please select a problem instance first.");
        return;
      }
      var progress = new Progress();
      progress.ProgressStateChanged += OkbDownloadProgressOnStateChanged;
      Content.UpdateKnowledgeBaseAsync(progress);
      MainFormManager.GetMainForm<HeuristicLab.MainForm.WindowsForms.MainForm>().AddOperationProgressToView(progressPanel, progress);
      progressPanel.Visible = true;
      SetEnabledStateOfControls();
    }

    private void OkbDownloadProgressOnStateChanged(object sender, EventArgs e) {
      if (InvokeRequired) { Invoke((Action<object, EventArgs>)OkbDownloadProgressOnStateChanged, sender, e); return; }
      var progress = (IProgress)sender;
      okbDownloadInProgress = progress.ProgressState == ProgressState.Started;
      SetEnabledStateOfControls();
      if (!okbDownloadInProgress) {
        progressPanel.Visible = false;
        progress.ProgressStateChanged -= OkbDownloadProgressOnStateChanged;
      }
    }

    private void AlgorithmStartButtonOnClick(object sender, EventArgs e) {
      if (suggestedInstancesComboBox.SelectedIndex >= 0)
        Content.StartAlgorithmAsync(suggestedInstancesComboBox.SelectedIndex);
    }

    private void AlgorithmCloneButtonOnClick(object sender, EventArgs e) {
      if (suggestedInstancesComboBox.SelectedIndex >= 0)
        MainFormManager.MainForm.ShowContent((IAlgorithm)Content.SuggestedInstances[suggestedInstancesComboBox.SelectedIndex].Clone());
    }

    private void ProjectionComboBoxOnSelectedIndexChanged(object sender, EventArgs e) {
      if (InvokeRequired) { Invoke((Action<object, EventArgs>)ProjectionComboBoxOnSelectedIndexChanged, sender, e); return; }
      if (projectionComboBox.SelectedIndex < 0) return;
      var projection = (string)projectionComboBox.SelectedItem;
      var instancesSeries = instanceMapChart.Series["InstancesSeries"];
      var currentInstanceSeries = instanceMapChart.Series["CurrentInstanceSeries"];

      instancesSeries.Points.Clear();
      currentInstanceSeries.Points.Clear();

      foreach (var run in Content.ProblemInstances) {
        var xKey = "Projection." + projection + ".X";
        var yKey = "Projection." + projection + ".Y";
        if (!run.Results.ContainsKey(xKey) || !run.Results.ContainsKey(yKey)
          || !(run.Results[xKey] is Data.DoubleValue) || !(run.Results[yKey] is Data.DoubleValue)) continue;
        var x = ((Data.DoubleValue)run.Results[xKey]).Value;
        var y = ((Data.DoubleValue)run.Results[yKey]).Value;
        var dataPoint = new DataPoint(x, y) {
          Label = run.Name
        };
        instancesSeries.Points.Add(dataPoint);
      }

      var curPoint = Content.ProjectCurrentInstance(projection);
      if (curPoint != null) {
        var dp = new DataPoint(curPoint.Item1, curPoint.Item2) {
          Label = Content.Problem.Problem.Name
        };
        currentInstanceSeries.Points.Add(dp);
      }
    }

    private void SuggestedInstancesComboBoxOnSelectedIndexChanged(object sender, EventArgs e) {
      if (InvokeRequired) { Invoke((Action<object, EventArgs>)SuggestedInstancesComboBoxOnSelectedIndexChanged, sender, e); return; }
      if (suggestedInstancesComboBox.SelectedIndex >= 0) {
        solverParametersView.Content = Content.SuggestedInstances[suggestedInstancesComboBox.SelectedIndex].Parameters;
      } else solverParametersView.Content = null;
      SetEnabledStateOfControls();
    }

    private void SimilarityComboBoxOnSelectedIndexChanged(object sender, EventArgs e) {
      if (InvokeRequired) { Invoke((Action<object, EventArgs>)SimilarityComboBoxOnSelectedIndexChanged, sender, e); return; }
      var calculator = (ISolutionSimilarityCalculator)similarityComboBox.SelectedItem;
      if (calculator != null) {
        calculator.SolutionVariableName = (string)solutionNameComboBox.SelectedItem;
        calculator.QualityVariableName = Content.Problem.Problem.Evaluator.QualityParameter.ActualName;
      }
      UpdateSolutionDiversityAnalysis(calculator);
      UpdateSolutionFdcAnalysis(calculator);
      UpdateSolutionLengthScaleAnalysis(calculator);
      UpdateSolutionNetworkAnalysis(calculator);
    }
    #endregion
    #endregion

    #region Control Configuration
    private void UpdateSuggestedInstancesCombobox() {
      var prevSelection = (IAlgorithm)suggestedInstancesComboBox.SelectedItem;
      var prevNewIndex = -1;
      suggestedInstancesComboBox.Items.Clear();
      if (Content == null) return;

      for (var i = 0; i < Content.SuggestedInstances.Count; i++) {
        suggestedInstancesComboBox.Items.Add(Content.SuggestedInstances[i]);
        if (prevSelection == null || Content.SuggestedInstances[i].Name == prevSelection.Name)
          prevNewIndex = prevSelection == null ? 0 : i;
      }
      if (prevNewIndex >= 0) {
        suggestedInstancesComboBox.SelectedIndex = prevNewIndex;
      }
    }

    private void UpdateSimilarityCalculators() {
      var selected = (ISolutionSimilarityCalculator)(similarityComboBox.SelectedIndex >= 0 ? similarityComboBox.SelectedItem : null);
      similarityComboBox.Items.Clear();

      if (Content == null || Content.Problem == null) return;

      foreach (var calc in Content.Problem.Operators.OfType<ISolutionSimilarityCalculator>()) {
        similarityComboBox.Items.Add(calc);
        if (selected != null && calc.ItemName == selected.ItemName) similarityComboBox.SelectedItem = calc;
      }
      if (selected == null && similarityComboBox.Items.Count > 0)
        similarityComboBox.SelectedIndex = 0;
    }

    private void UpdateNamesComboboxes() {
      var selectedSolutionName = solutionNameComboBox.SelectedIndex >= 0 ? (string)solutionNameComboBox.SelectedItem : string.Empty;

      solutionNameComboBox.Items.Clear();
      if (Content == null) return;
      var solutionNames = Content.Problem.Solutions.Select(x => x.Solution).OfType<IScope>().SelectMany(x => x.Variables);

      foreach (var sn in solutionNames.GroupBy(x => x.Name).OrderBy(x => x.Key)) {
        solutionNameComboBox.Items.Add(sn.Key);
        // either it was previously selected, or the variable value is defined in the HeuristicLab.Encodings sub-namespace
        if (sn.Key == selectedSolutionName || (string.IsNullOrEmpty(selectedSolutionName) && sn.All(x => x.Value != null && x.Value.GetType().FullName.StartsWith("HeuristicLab.Encodings."))))
          solutionNameComboBox.SelectedItem = sn.Key;
      }
    }
    
    private void UpdateProjectionComboBox() {
      projectionComboBox.Items.Clear();
      var projStrings = Content.ProblemInstances
        .SelectMany(x => x.Results.Where(y => Regex.IsMatch(y.Key, "^Projection[.].*[.][XY]$")))
        .Select(x => Regex.Match(x.Key, "Projection[.](?<g>.*)[.][XY]").Groups["g"].Value)
        .Distinct();
      foreach (var str in projStrings) {
        projectionComboBox.Items.Add(str);
      }
    }
    #endregion

    #region Visualization
    #region Solution Visualization
    public void UpdateSolutionVisualization() {
      if (InvokeRequired) { Invoke((Action)UpdateSolutionVisualization); return;  }
      var qualityName = Content.Problem.Problem.Evaluator.QualityParameter.ActualName;
      UpdateSolutionQualityAnalysis(qualityName);
      if (similarityComboBox.SelectedIndex >= 0) {
        var calculator = (ISolutionSimilarityCalculator)similarityComboBox.SelectedItem;
        UpdateSolutionDiversityAnalysis(calculator);
        UpdateSolutionFdcAnalysis(calculator);
        UpdateSolutionLengthScaleAnalysis(calculator);
        UpdateSolutionNetworkAnalysis(calculator);
      } else {
        solutionsDiversityViewHost.Content = null;
        solutionsFdcViewHost.Content = null;
        solutionsLengthScaleViewHost.Content = null;
        solutionsNetworkChart.Series.First().Points.Clear();
      }
    }
    private void UpdateSolutionQualityAnalysis(string qualityName) {
      var dt = solutionsQualityViewHost.Content as DataTable;
      if (dt == null) {
        dt = QualityDistributionAnalyzer.PrepareTable(qualityName);
        dt.VisualProperties.Title = "Quality Distribution";
        solutionsQualityViewHost.Content = dt;
      }
      QualityDistributionAnalyzer.UpdateTable(dt, Content.Problem.Solutions.Select(x => x.Solution).OfType<IScope>().Where(x => x.Variables.ContainsKey(qualityName)).Select(x => x.Variables[qualityName].Value).OfType<DoubleValue>().Select(x => x.Value));
    }

    private void UpdateSolutionDiversityAnalysis(ISolutionSimilarityCalculator calculator) {
      try {
        solutionsDiversityViewHost.Content = null;
        var similarities = new double[Content.Problem.Solutions.Count, Content.Problem.Solutions.Count];
        for (int i = 0; i < Content.Problem.Solutions.Count; i++) {
          for (int j = 0; j < Content.Problem.Solutions.Count; j++)
            similarities[i, j] = calculator.CalculateSolutionSimilarity((IScope)Content.Problem.Solutions[i].Solution, (IScope)Content.Problem.Solutions[j].Solution);
        }
        var hm = new HeatMap(similarities, "Solution Similarities", 0.0, 1.0);
        solutionsDiversityViewHost.Content = hm;
      } catch { }
    }

    private void UpdateSolutionFdcAnalysis(ISolutionSimilarityCalculator calculator) {
      try {
        solutionsFdcViewHost.Content = null;
        var points = new List<Point2D<double>>();
        for (int i = 0; i < Content.Problem.Solutions.Count; i++) {
          for (int j = 0; j < Content.Problem.Solutions.Count; j++) {
            if (i == j) continue;
            var qDiff = Math.Abs(((DoubleValue)((IScope)Content.Problem.Solutions[i].Solution).Variables[calculator.QualityVariableName].Value).Value
              - ((DoubleValue)((IScope)Content.Problem.Solutions[j].Solution).Variables[calculator.QualityVariableName].Value).Value);
            points.Add(new Point2D<double>(qDiff, 1.0 - calculator.CalculateSolutionSimilarity((IScope)Content.Problem.Solutions[i].Solution, (IScope)Content.Problem.Solutions[j].Solution)));
          }
        }
        var splot = new ScatterPlot("Fitness-Distance", "");
        splot.VisualProperties.XAxisTitle = "Absolute Fitness Difference";
        splot.VisualProperties.XAxisMinimumFixedValue = 0.0;
        splot.VisualProperties.YAxisTitle = "Solution Distance";
        splot.VisualProperties.YAxisMinimumFixedValue = 0.0;
        splot.VisualProperties.YAxisMaximumFixedValue = 1.0;
        var row = new ScatterPlotDataRow("Fdc", "", points);
        row.VisualProperties.PointSize = 7;
        splot.Rows.Add(row);
        solutionsFdcViewHost.Content = splot;
      } catch { }
    }

    private void UpdateSolutionLengthScaleAnalysis(ISolutionSimilarityCalculator calculator) {
      try {
        solutionsLengthScaleViewHost.Content = null;
        var dt = solutionsLengthScaleViewHost.Content as DataTable;
        if (dt == null) {
          dt = QualityDistributionAnalyzer.PrepareTable("Length Scale");
          solutionsLengthScaleViewHost.Content = dt;
        }

        QualityDistributionAnalyzer.UpdateTable(dt, CalculateLengthScale(calculator));
      } catch { }
    }

    private IEnumerable<double> CalculateLengthScale(ISolutionSimilarityCalculator calculator) {
      for (int i = 0; i < Content.Problem.Solutions.Count; i++) {
        for (int j = 0; j < Content.Problem.Solutions.Count; j++) {
          if (i == j) continue;
          var sim = calculator.CalculateSolutionSimilarity((IScope)Content.Problem.Solutions[i].Solution, (IScope)Content.Problem.Solutions[j].Solution);
          if (sim.IsAlmost(0)) continue;
          var qDiff = Math.Abs(((DoubleValue)((IScope)Content.Problem.Solutions[i].Solution).Variables[calculator.QualityVariableName].Value).Value
            - ((DoubleValue)((IScope)Content.Problem.Solutions[j].Solution).Variables[calculator.QualityVariableName].Value).Value);
          yield return qDiff / sim;
        }
      }
    }

    private void UpdateSolutionNetworkAnalysis(ISolutionSimilarityCalculator calculator) {
      try {
        var series = solutionsNetworkChart.Series["SolutionSeries"];
        series.Points.Clear();
        var dissimilarities = new DoubleMatrix(Content.Problem.Solutions.Count, Content.Problem.Solutions.Count);
        for (int i = 0; i < Content.Problem.Solutions.Count; i++) {
          for (int j = 0; j < Content.Problem.Solutions.Count; j++) {
            if (i == j) continue;
            dissimilarities[i, j] = 1.0 - calculator.CalculateSolutionSimilarity((IScope)Content.Problem.Solutions[i].Solution, (IScope)Content.Problem.Solutions[j].Solution);
          }
        }
        var coords = MultidimensionalScaling.KruskalShepard(dissimilarities);
        for (var i = 0; i < coords.Rows; i++) {
          var quality = ((DoubleValue)((IScope)Content.Problem.Solutions[i].Solution).Variables[calculator.QualityVariableName].Value).Value;
          series.Points.Add(new DataPoint() {
            Name = (i + 1).ToString(),
            XValue = coords[i, 0],
            YValues = new[] { coords[i, 1], quality },
            Label = i + ": " + quality
          });
        }
      } catch { }
    }
    #endregion
    #endregion
  }
}
